import pandas as pd

from autogluon.common.features.types import R_CATEGORY, R_FLOAT, R_INT, S_TEXT


class VWFeaturesConverter:
    """
    Converts features in PandasDataFrame to VW format
    Ref: https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Input-format
    """

    PIPE = "|"
    SPACE = " "

    # TODO: Add support for different namespaces

    def convert_features_to_vw_format(self, X, feature_metadata) -> pd.Series:
        """
        Converts features to VW format.
        :param X: features
        :param feature_metadata: schema of X
        :return: Returns a series of features converted to VW format
        """

        X_out: pd.Series = None

        for feature in feature_metadata:
            raw_feature, special_feature = feature_metadata[feature]
            if X_out is None:
                X_out = (
                    self.PIPE
                    + self.SPACE
                    + self.__generate_namespace_based_on_ml_type(X[feature], raw_feature, special_feature, feature).astype("str")
                    + self.SPACE
                )
            else:
                X_out += (
                    "" + self.SPACE + self.__generate_namespace_based_on_ml_type(X[feature], raw_feature, special_feature, feature).astype("str") + self.SPACE
                )
        return X_out

    def __generate_namespace_based_on_ml_type(self, input_series, raw_feature, special_feature, feature_name=None):
        """
        Based on the type of feature, preprocess/sanify these features so that it is in VW format
        Only use raw text, numeric integer, numeric decimals, and category
        Ref: https://github.com/autogluon/autogluon/blob/master/common/src/autogluon/common/features/types.py

        :param input_series: A single feature as Pandas Series
        :param raw_feature: Raw feature Type
        :param special_feature: Special Feature Type
        :param feature_name: Column Name of this feature
        :return: Preprocessed Feature as a Pandas Series
        """
        if S_TEXT in special_feature:
            return input_series.apply(self.__preprocess_text)
        elif raw_feature in [R_INT, R_FLOAT]:
            return input_series.apply(self.__numeric_namespace_generator, args=(feature_name,))
        elif raw_feature == R_CATEGORY:
            return input_series.apply(self.__categorical_namespace_generator, args=(feature_name,))
        else:
            raise ValueError(
                f"Received unsupported raw_feature_type '{str(raw_feature)}' special_feature_type '{str(special_feature)}'"
                f" for feature '{feature_name}' for class {self.__class__.__name__}."
            )

    def __preprocess_text(self, s) -> str:
        if pd.isnull(s):
            return ""
        s = " ".join(str(s).split())
        # Added split to remove tabs spaces since tab is used as separator
        text = self.__sanify(s)
        return text

    def __sanify(self, s) -> str:
        """
        The sanify is performed because : and | are reserved by vowpal wabbit for distinguishing namespaces and numeric
        data
        @param s: input string
        @returns string
        """
        return str(s).replace(":", ";").replace("|", "/")

    def __numeric_namespace_generator(self, feature, feature_name) -> str:
        if pd.isnull(feature):
            return ""
        return feature_name + ":" + str(feature)

    def __categorical_namespace_generator(self, feature, feature_name) -> str:
        if pd.isnull(feature):
            return ""
        else:
            feature = str(feature).replace(" ", "_")
            return feature_name + "=" + self.__sanify(feature)



import torch
from torch.optim.lr_scheduler import ReduceLROnPlateau


def get_scheduler(hyperparams: dict, optimizer: torch.optim.Optimizer):

    if hyperparams['lr_scheduler']:      
        scheduler = ReduceLROnPlateau(
            optimizer, 
            patience=hyperparams['lr_scheduler_patience'], 
            min_lr=0, 
            factor=0.2
        )
    else:
        scheduler = ReduceLROnPlateau(
            optimizer, 
            patience=10000000, 
            min_lr=0, 
            factor=0.2
        )

    return scheduler



import logging

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_selection import SelectKBest, f_classif, f_regression
from sklearn.preprocessing import QuantileTransformer

from ..core.enums import Task

logger = logging.getLogger(__name__)


class Preprocessor(TransformerMixin, BaseEstimator):
    """
    This class is used to preprocess the data before it is pushed through the model.
    The preprocessor assures that the data has the right shape and is normalized,
    This way the model always gets the same input distribution, 
    no matter whether the input data is synthetic or real.

    """

    def __init__(
            self, 
            max_features: int,
            use_quantile_transformer: bool,
            use_feature_count_scaling: bool,
            task: Task,
        ):

        self.max_features = max_features
        self.use_quantile_transformer = use_quantile_transformer
        self.use_feature_count_scaling = use_feature_count_scaling
        self.task = task
    
    def fit(self, X: np.ndarray, y: np.ndarray):

        self.compute_pre_nan_mean(X)
        X = self.impute_nan_features_with_mean(X)

        self.determine_which_features_are_singular(X)
        X = self.cutoff_singular_features(X, self.singular_features)

        self.determine_which_features_to_select(X, y)
        X = self.select_features(X)

        if self.use_quantile_transformer:
            n_obs, n_features = X.shape
            n_quantiles = min(n_obs, 1000)
            self.quantile_transformer = QuantileTransformer(n_quantiles=n_quantiles, output_distribution='normal')
            X = self.quantile_transformer.fit_transform(X)
        
        self.mean, self.std = self.calc_mean_std(X)
        X = self.normalize_by_mean_std(X, self.mean, self.std)

        assert np.isnan(X).sum() == 0, "There are NaNs in the data after preprocessing"

        return self
    

    def transform(self, X: np.ndarray):

        X = self.cutoff_singular_features(X, self.singular_features)
        X = self.impute_nan_features_with_mean(X)
        X = self.select_features(X)

        if self.use_quantile_transformer:
            X = self.quantile_transformer.transform(X)
        
        X = self.normalize_by_mean_std(X, self.mean, self.std)

        if self.use_feature_count_scaling:
            X = self.normalize_by_feature_count(X, self.max_features)

        X = self.extend_feature_dim_to_max_features(X, self.max_features)

        assert np.isnan(X).sum() == 0, "There are NaNs in the data after preprocessing"

        return X
        

    def determine_which_features_are_singular(self, x: np.ndarray) -> None:

        self.singular_features = np.array([ len(np.unique(x_col)) for x_col in x.T ]) == 1

    def determine_which_features_to_select(self, x: np.ndarray, y: np.ndarray) -> None:

        if x.shape[1] > self.max_features:
            logger.info(f"A maximum of {self.max_features} features are allowed, but the dataset has {x.shape[1]} features. A subset of {self.max_features} are selected using SelectKBest")

            if self.task == Task.CLASSIFICATION:
                self.select_k_best = SelectKBest(k=self.max_features, score_func=f_classif)
            else:  # Task.REGRESSION
                self.select_k_best = SelectKBest(k=self.max_features, score_func=f_regression)
            self.select_k_best.fit(x, y)

    def compute_pre_nan_mean(self, x: np.ndarray) -> None:
        """
        Computes the mean of the data before the NaNs are imputed
        """
        self.pre_nan_mean = np.nanmean(x, axis=0)


    def impute_nan_features_with_mean(self, x: np.ndarray) -> np.ndarray:

        inds = np.where(np.isnan(x))
        x[inds] = np.take(self.pre_nan_mean, inds[1])
        return x

    
    def select_features(self, x: np.ndarray) -> np.ndarray:

        if x.shape[1] > self.max_features:
            x = self.select_k_best.transform(x)

        return x
    

    def cutoff_singular_features(self, x: np.ndarray, singular_features: np.ndarray) -> np.ndarray:

        if singular_features.any():
            x = x[:, ~singular_features]

        return x


    def calc_mean_std(self, x: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        """
        Calculates the mean and std of the training data
        """
        mean = x.mean(axis=0)
        std = x.std(axis=0)
        return mean, std
    

    def normalize_by_mean_std(self, x: np.ndarray, mean: np.ndarray, std: np.ndarray) -> np.ndarray:
        """
        Normalizes the data by the mean and std
        """

        x = (x - mean) / std
        return x


    def normalize_by_feature_count(self, x: np.ndarray, max_features) -> np.ndarray:
        """
        An interesting way of normalization by the tabPFN paper
        """

        x = x * max_features / x.shape[1]
        return x



    def extend_feature_dim_to_max_features(self, x: np.ndarray, max_features) -> np.ndarray:
        """
        Increases the number of features to the number of features the model has been trained on
        """
        added_zeros = np.zeros((x.shape[0], max_features - x.shape[1]), dtype=np.float32)
        x = np.concatenate([x, added_zeros], axis=1)
        return x
    


    
